/*
*  FileMan.h
*  Copyright  1999-2006 by Richard Christman
*  ALL RIGHTS RESERVED
*
*  This source file is released for peer review purposes only. It is not public domain. 
*  Redistribution either in part or whole is forbidden. The sole source for QuickSilver 
*  distribution currently is http://quicksilvermail.net . You have permission to compile
*  this source for your own use.
*
*  Richard
*  --
*  R.Christman
*  quicksilver@quicksilvermail.net
*  http://quicksilvermail.net
*
*/
//---------------------------------------------------------------------------
#ifndef Fileman1H
#define Fileman1H

#include "Global.h"
//---------------------------------------------------------------------------
#define FLNIL -1L
// header = 38 chars
// "BS\150\209\200\194\254\228\178\139\204\182"
// "qs\002\009\002\004\000\008\000\007\002\002\000\000\001\001\001\002\005"
//   "\150\209\200\194\254\228\178\139\204\182
//#define FILEHEADER     "\130\137\130\132\128\136\128\135\130\182"
#define FILEHEADER "qsdb1xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\n"
#define FILEHEADE1 "bsdbxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\n"
#define AppDatFile AnsiString(AppDat)
#define PS AnsiString("\xff")
#define INTSIZE sizeof(int)
#define PTRSIZE sizeof(void*)
#define OFSSIZE sizeof(TFLoffset)
#define SPCSIZE sizeof(TFLspace)
#define TYPSIZE 4

class TDataBuf;
class TFreeList;
typedef long TFLoffset;
typedef long TFLspace;
const TFLspace sizeofHeader = SPCSIZE;

AnsiString __fastcall GetAppIni(AnsiString &key, AnsiString &defvalue, AnsiString &datafile=AppDatFile);
int __fastcall PutAppIni(AnsiString &key, AnsiString &Value, AnsiString &datafile=AppDatFile);
int __fastcall PutAppData(AnsiString &path, TStrings *items, AnsiString &datafile=AppDatFile);
int __fastcall GetAppData(TStrings *items, AnsiString &path, AnsiString &datafile=AppDatFile);
int __fastcall GetAppDataMainIndex(TStrings *items, AnsiString &datafile=AppDatFile);
int __fastcall GetAppBinary(AnsiString &path, TDataBuf *buf, AnsiString &datafile=AppDatFile);
int __fastcall PutAppBinary(AnsiString &path, TDataBuf *buf, AnsiString &datafile=AppDatFile);
int __fastcall PutAppBinaryData(AnsiString &path, char *data, int length, AnsiString &datafile=AppDatFile);
bool __fastcall AppDataExists(AnsiString &path, TStrings *data=NULL, AnsiString &datafile=AppDatFile);
bool __fastcall AppBinaryExists(AnsiString &path, TDataBuf *data=NULL, AnsiString &datafile=AppDatFile);
int __fastcall RenameAppIndex(AnsiString &Path, AnsiString &Newname, AnsiString &datafile=AppDatFile);
int __fastcall RenameAppItem(AnsiString &Path, AnsiString &Newname, AnsiString &datafile=AppDatFile);

class TFileManager : public TObject {
private:
  TStream *stream;
  TList *stats;
  TFreeList *freeList;	    // Free list block in memory.
  TFLoffset startOfData;	// Offset to first data in file.
  TFLoffset endOfData;		// Offset to last data in file.
  bool delStream;
public:
  __property int dbstatus   =  { read=HiStatus };
  __property int histatus   =  { read=HiStatus };
  __property int lostatus   =  { read=LoStatus };

  int block;
  AnsiString errormessage,filename;                                       // create if not exist
  __fastcall virtual TFileManager(const AnsiString &Filename, bool CreateIfNeed=true);
  __fastcall virtual TFileManager(TStream *Stream);
  __fastcall ~TFileManager();
  void __fastcall Init(); // does all construction except stream
  TFLoffset __fastcall start() const { return startOfData; }
  TFLoffset __fastcall endData() const  { return endOfData; }

  void __fastcall status(int status);
  bool __fastcall GetStatus(int &status, AnsiString &statmsg);
  int __fastcall HiStatus();
  int __fastcall LoStatus();

  TFLoffset __fastcall Allocate(TFLspace spc);
  //TFLoffset __fastcall AllocateInFreeList(TFLspace spc);
  bool __fastcall AllocateInFreeList(TFLoffset &offset, TFLspace spaceRequired);
  bool __fastcall AllocateAtEnd(TFLoffset &offset, TFLspace spaceRequired);
  TFLoffset __fastcall AllocateAtEnd(TFLspace spc);
  void __fastcall Deallocate(TFLoffset ofs);

  bool __fastcall WriteFreeList();
  bool __fastcall ReadFreeList();
  bool __fastcall AddToFreeList(TFLoffset ofs, TFLspace space);

  bool __fastcall SeekTo(const TFLoffset ofs);
  bool __fastcall WriteDat(void *data, long size);
  bool __fastcall WriteOfs(const TFLoffset ofs, void *data, long size);
  bool __fastcall WriteHeader();
  bool __fastcall WriteStartAndEnd();

  bool __fastcall ReadStartAndEnd();
  bool __fastcall ReadHeader();
  bool __fastcall ReadDat(void *data, long size);
  bool __fastcall ReadOfs(const TFLoffset ofs, void *data, long size);

  AnsiString __fastcall GetErrorMessage(int ErrNo);
  bool __fastcall IsError(AnsiString ErrMsg);
  bool __fastcall FatalError();
  bool __fastcall HasError();
};

const char CLEARTEXT = 0;
const char COMPRESSED = 1;
const char ENCRYPTED = 2;
const char ISINDEX = 1;
const char ISDATA = 2;

class TDataBuf : public TObject {
public:
   int status;
   char* data;
   bool sensitive;
   char type[TYPSIZE];
   TFLspace length,orglength;
   virtual __fastcall TDataBuf(const AnsiString &Data);
   virtual __fastcall TDataBuf(const char *Data=NULL, TFLspace Length=0);
   __fastcall ~TDataBuf();
   bool __fastcall LoadFromFile(const AnsiString &Filename);
   bool __fastcall SaveToFile(const AnsiString &Filename);
   void __fastcall SetData(const char *Data, TFLspace Length); // sets 'data' and 'length' only
   void __fastcall Set(const char *Data, TFLspace Length);     // resets all
   int  __fastcall Encrypt(const AnsiString &password);
   int  __fastcall Decrypt(const AnsiString &password);
   bool __fastcall Encrypted();
   bool __fastcall Compressed();
   int  __fastcall Compress();
   int  __fastcall Decompress();
   bool __fastcall IsIndex();
   bool __fastcall IsData();
   bool __fastcall IsEmpty();
   void __fastcall Clear();
   AnsiString __fastcall str();
   char* __fastcall c_str();
};

typedef TDataBuf TFileBuf;
typedef TDataBuf TBuffer;

class TFreeList : public TObject {
private:
   TList *offset;
   TList *space;
public:
   int status;
   TFLoffset lastOffset; // assigned but not used
   TFLspace lastSpace;   //    ""    ""  ""   ""
   __fastcall TFreeList();
   __fastcall ~TFreeList();
   bool __fastcall AddFreeSpace(TFLoffset, TFLspace);
   TFLoffset __fastcall Allocate(TFLspace);
   void __fastcall AttemptMerge(int);
   bool __fastcall Coalesce(TFLoffset, TFLspace);
   bool __fastcall EnterFreeSpace(TFLoffset, TFLspace);
   TFLoffset __fastcall getOffset(int i) { return (TFLoffset)offset->Items[i]; }
   TFLspace __fastcall getSpace(int i) { return (TFLspace)space->Items[i]; }
   bool __fastcall LoadFromStream(TStream *);
   bool __fastcall SaveToStream(TStream *);
   int __fastcall Count();
};

class TFMindex : public TStringList {
public:
   AnsiString name;
   TFLoffset offset;
   TFMindex *parent;
   TFileManager *fm;
   virtual __fastcall TFMindex(TFileManager *Fm, AnsiString Name="", TFMindex *Parent=NULL);
   TFLoffset __fastcall Save();
   void __fastcall ShowIndex();
   char __fastcall GetType(AnsiString &key);
   TFLspace  __fastcall GetSize(AnsiString &key);
   TFLoffset __fastcall GetOffset(AnsiString &key);
   TFLoffset __fastcall SetOffset(AnsiString &key, TFLoffset Offset); // <-not to disk
   TFLoffset __fastcall PutOffset(AnsiString &key, TFLoffset Offset); // <-to disk
   char* __fastcall GetBinary(AnsiString &key, TDataBuf* buf);
   // functions below just write data and update index in memory, but don't
   // save index, or in the case of writeindex, don't cascade to parent..
   virtual TFLoffset __fastcall PutBinary(AnsiString &key, TDataBuf *buf, bool SaveIndex=true);
   //virtual TFLoffset __fastcall PutBinary(AnsiString &key, TMemoryStream *buf, bool SaveIndex=true);
   TFLoffset __fastcall WriteIndex();
};

// FileManager
#define DB_OK				    0
#define DB_FATALERR             1
#define DB_NODBERR				10
#define DB_SEEKERR              11
#define DB_SEEKEXC              12
#define DB_READERR              13
#define DB_READEXC              14
#define DB_WRITEERR             15
#define DB_WRITEEXC             16
#define DB_OPENDBEXC            17
#define DB_INITVIRGINDBERR      18
#define DB_INITEXISTDBERR       19
#define DB_READHEADERR          20
#define DB_READHEADEXC          21
#define DB_INVALIDHEADERR       22
#define DB_READSODERR           23
#define DB_READSODEXC           24
#define DB_SEEKFREEERR          25
#define DB_READFREEERR          26
#define DB_READFREEEXC          27
#define DB_WRITEHEADERR         28
#define DB_WRITEHEADEXC         29
#define DB_WRITEFREEERR         30
#define DB_WRITEFREEEXC         31
#define DB_WRITESODERR          32
#define DB_WRITESODEXC          33
#define DB_ALLOCERR             34
#define DB_WRITEALLOCERR        35
#define DB_ALLOCEXC             36
#define DB_WRITEALLOCFREEERR    37
#define DB_ALLOCFREEERR         38
#define DB_ALLOCFREEEXC         39
#define DB_WRITEALLOCENDERR     40
#define DB_WRITEALLOCSODERR     41
#define DB_ALLOCATENDEXC        42
#define DB_DEALOCREADSORERR     43
#define DB_DEALOCREADFREEERR    44
#define DB_DEALLOCFREEERR       45
#define DB_ADDFREEWRITEERR      46
#define DB_ADDFREEENTERERR      47
#define DB_GETBINARYERR         48
#define DB_GETBINALLOCEXC       49
#define DB_PUTBINARYERR         50
#define DB_OPENINDEXERR         51
#define DB_OPENINDEXEXC         52
#define DB_GETDATSIZEERR        53
#define DB_GETDATTYPEERR        54
#define DB_WRITEINDEXERR        55
#define DB_BADBUFERR            56
#define DB_INITINDEXERR         57
#define DB_ZIPDEFLERR			58
#define DB_ZIPINFLERR           59
#define DB_ALLOCINDEXEXC        60
#define DB_WRITEINDEXOFSERR     61
#define DB_READINDEXERR         62
#define DB_READINDEXEXC         63
#define DB_SAVEINDEXERR         64

// FileBuf
#define FB_OK					0
#define FB_NOFILEERR            70
#define FB_OPENREADFILEEXC      71
#define FB_OPENWRITEFILEEXC     72
#define FB_READFILEEXC          73
#define FB_WRITEFILEEXC         74
#define FB_ALLOCBUFERR          75
#define FB_ALLOCBUFEXC          76
#define FB_ENCRYPTEXC			77
#define FB_DECRYPTEXC           78
#define FB_ZIPINFLERR           79
#define FB_ZIPINFLEXC           70
#define FB_ZIPDEFLERR           71
#define FB_ZIPDEFLEXC           72

// FreeList
#define FL_OK					0
#define FL_READERR              80
#define FL_READEXC              81
#define FL_WRITEERR             82
#define FL_WRITEEXC             83

extern bool DBFATALerror;
extern bool FileManWait;
//---------------------------------------------------------------------------
#endif
