/*
*  WipeDlgs.cpp
*  Copyright  1999-2006 by Richard Christman
*  ALL RIGHTS RESERVED
*
*  This source file is released for peer review purposes only. It is not public domain. 
*  Redistribution either in part or whole is forbidden. The sole source for QuickSilver 
*  distribution currently is http://quicksilvermail.net . You have permission to compile
*  this source for your own use.
*
*  Richard
*  --
*  R.Christman
*  quicksilver@quicksilvermail.net
*  http://quicksilvermail.net
*
*/
//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#include <dir.h>
#pragma hdrstop

#include "WipeDlgs.h"
#include "fileman.h"
#include "Global.h"
#include "Fmxutils.h"
#include "mixmaster.h"
#include <fcntl.h>
#include <io.h>
#include <dos.h>
#include <errno.h>
//---------------------------------------------------------------------------
#pragma resource "*.dfm"
TWipeDlg *WipeDlg;
AnsiString WipeErrorStr;
char *sessionkey=NULL;
//---------------------------------------------------------------------------
AnsiString __fastcall GetWipeErrorStr();
int __fastcall GetWipeBuffer(struct isaac_state *state, int pass, int buffer[256],
   int *patterns);
//---------------------------------------------------------------------------
// Begin defines, structures, prototypes from Bob Jenkins' ISAAC generator
// (http://ourworld.compuserve.com/homepages/bob_jenkins/isaacafa.htm)
#define ISAAC_LOG 8
#define ISAAC_SIZE 256
#define ISAAC_BYTES (ISAAC_SIZE * sizeof(int))
// These are the disk-wiping patterns.  They are repeating patterns of
// 1, 2, 3 and 4 bits, along with some random passes.  See Peter Gutmann's
// paper, "Secure Deletion of Data from Magnetic and Solid-State Memory",
// first published in the Sixth USENIX Security Symposium Proceedings,
// San Jose, California, July 22-25, 1996 for the gory details why.
// (http://www.cs.auckland.ac.nz/~pgut001/pubs/secure_del.html)
const int patterns[] = {
	-1, // Random
	-1, // Random
	0x555, 0xAAA,
	0x249, 0x492, 0x924,
	0x000, 0xFFF,
	0x111, 0x222, 0x333, 0x444, 0x666, 0x777,
	0x888, 0x999, 0xBBB, 0xCCC, 0xDDD, 0xEEE,
	0xDB6, 0xB6D, 0x6DB,
	-1, // Random
	-1  // Random
};
#define NUMPATTERNS (sizeof(patterns)/sizeof(patterns[0]))
typedef struct isaac_state isaac_state;
struct isaac_state
{
	 int mm[ISAAC_SIZE]; // state
	 int aa, bb, cc;
};
static void isaac_seed(struct isaac_state *state);
static void filltable(struct isaac_state *state, int buf[ISAAC_SIZE], int type);
static void shuffle(int *array, int num, struct isaac_state *s);
static void tweakshuffle(int *array, int num);
// End Bob Jenkins declarations
//---------------------------------------------------------------------------
char* __fastcall GetSessionKey()
{
   if (sessionkey) return sessionkey;

   struct isaac_state state;
   sessionkey = new char[ISAAC_BYTES+1];
   RAND_bytes((byte*)sessionkey,21); byte rndbte = sessionkey[21];
   isaac_seed(&state); filltable(&state,(int*)sessionkey,-1);
   for (int i=0; i<ISAAC_BYTES; i++) if (sessionkey[i]==0) sessionkey[i] = rndbte;
   sessionkey[ISAAC_BYTES] = '\0';
   if (ISDESIGNING)
   {
      AnsiString str,key=FormatDateTime("mm-dd-yy hh:nn",Now());
      AnsiString seskeystxt = AppDir+"0seskeys.txt";
      TStringList *seskeys = new TStringList();
      if (FileExists(seskeystxt)) seskeys->LoadFromFile(seskeystxt);
      for (int i=0; i<128; i++)
      {
         unsigned char x = sessionkey[i]; str = s(x);
         while (str.Length()<4) str = " "+str;
         key += str;
      }
      seskeys->Add(key);
      seskeys->SaveToFile(seskeystxt);
      delete seskeys;
   }
   return sessionkey;
}
//---------------------------------------------------------------------------
int __fastcall dosimplewipe(AnsiString &Filename, int Passes)
{
   if (!FileExists(Filename)) return WIPE_NOTEXISTING;

   DWORD dwSize,dwPageSize,dwNumPages;
   int workpatterns[NUMPATTERNS];
   SYSTEM_INFO sysinfo;
   HANDLE hFile,hMap;
   int buffer[256];
   LPSTR lpFile;

   // get pagesize we'll use up to one extra
   GetSystemInfo(&sysinfo);
   dwPageSize = sysinfo.dwPageSize;
   // open file to wipe
   if ((hFile=CreateFile(Filename.c_str(),
      GENERIC_READ|GENERIC_WRITE,0,NULL,OPEN_EXISTING,
      FILE_ATTRIBUTE_NORMAL|FILE_FLAG_WRITE_THROUGH,NULL))==INVALID_HANDLE_VALUE)
   {
      WipeErrorStr = GetWipeErrorStr();
      if (WipeErrorStr.IsEmpty()) WipeErrorStr = "Failed to open file for wipeing!";
      return WIPE_NOTOPEN;
   }
   // if the file is empty, no need to wipe
   if ((dwSize=GetFileSize(hFile,NULL))==0)
   {
      CloseHandle(hFile);
      return WIPE_OK;
   }
   // get pages required
   dwNumPages = (dwSize+dwPageSize-1)/dwPageSize;
   dwSize = dwNumPages * dwPageSize;
   // create a file mapping object for the wipe file
   if ((hMap=CreateFileMapping(hFile,NULL,PAGE_READWRITE,0,0,NULL))==0)
   {
      WipeErrorStr = GetWipeErrorStr();
      if (WipeErrorStr.IsEmpty()) WipeErrorStr = "Failed to create file mapping device!";
      CloseHandle(hFile);
      return WIPE_NOFILEMAP;
   }
   // map the file into view
   if ((lpFile=(char*)MapViewOfFile(hMap,FILE_MAP_WRITE,0,0,0))==NULL)
   {
      WipeErrorStr = GetWipeErrorStr();
      if (WipeErrorStr.IsEmpty()) WipeErrorStr = "Failed to map view of file!";
      CloseHandle(hFile);
      CloseHandle(hMap);
      return WIPE_NOMEMMAP;
   }
   // move patterns to working array
   for (int i=0; i<NUMPATTERNS; i++) workpatterns[i] = patterns[i];
   // initialize issac_state
   struct isaac_state state;
   isaac_seed(&state);
   if(Passes>3)
 	 {
	    // only wanna shake up the array
	    if(Passes > NUMPATTERNS) Passes = NUMPATTERNS;
      // Shuffle, leaving first & last pass alone
		 shuffle(workpatterns+1,(unsigned)Passes-2,&(state));
		 // Force a random pass to the middle
		 tweakshuffle(workpatterns+1,(unsigned)Passes-2);
   }
   // do the actual wiping passes
   for(int pass=0; pass<Passes; pass++)
   {
      GetWipeBuffer(&state,pass,buffer,workpatterns);
      for (unsigned count=0; count<dwSize; count+=1024)
         memcpy(lpFile+count,buffer,1024);
      FlushViewOfFile(lpFile,dwSize);
      FlushFileBuffers(hFile);
   }
   // done--so do some clean up
   UnmapViewOfFile(lpFile);
   CloseHandle(hMap);
   FlushFileBuffers(hFile);
   CloseHandle(hFile);
   return WIPE_OK;
}
//---------------------------------------------------------------------------
int __fastcall zapfile(AnsiString &filename, int passes)
{
   int err = simplewipe(filename,passes);
   DeleteFile(filename.c_str());
   return err;
}
//---------------------------------------------------------------------------
int __fastcall simplewipe(AnsiString &Filename, int Passes)
{
   int err=dosimplewipe(Filename,Passes);
   if (err!=WIPE_OK && err!=WIPE_NOTEXISTING)
      MsgBox(
         "Unable to wipe file!\nError: "+WipeErrorStr,
         "Wiping Error",
          MB_OK
      );
   return err;
}
//---------------------------------------------------------------------------
int __fastcall wipefile(AnsiString &Filename, int Passes)
{
   int err=WIPE_OK;
   AppProcessMessages(); Sleep(200);
   WipeDlg = new TWipeDlg(AppMainForm);
   WipeDlg->files->Text = Filename;
   if (WipeDlg->files->Count)
   {
      WipeDlg->passes = Passes; WipeDlg->ShowModal();
      if ((err=WipeDlg->retcode)!=WIPE_OK)
         MsgBox(
            "Unable to wipe file!\nError: "+WipeErrorStr,
            "Wiping Error",
             MB_OK
         );
   }
   delete WipeDlg;
   return err;
}
//---------------------------------------------------------------------------
int __fastcall wipefiles(AnsiString &Filename, int Passes)
{
   int err=WIPE_OK;
   WipeDlg = new TWipeDlg(AppMainForm);
   WipeDlg->files->Text = Filename;
   int i=0; while (i<WipeDlg->files->Count)
   {
      if (!FileExists(WipeDlg->files->Strings[i]))
         WipeDlg->files->Delete(i);
      else
         i++;
   }
   if (WipeDlg->files->Count)
   {
      WipeDlg->passes = Passes;
      WipeDlg->ShowModal();
      err=WipeDlg->retcode;
   }
   delete WipeDlg;
   return err;
}
//---------------------------------------------------------------------------
__fastcall TWipeDlg::TWipeDlg(TComponent* Owner)
	: TForm(Owner)
{
   Caption = "Wipe Files";
   files = new TStringList();
   retcode = WIPE_OK;
   passes = 7;
}
//---------------------------------------------------------------------------
void __fastcall TWipeDlg::FormDestroy(TObject *Sender)
{
   delete files;
}
//---------------------------------------------------------------------------
void __fastcall TWipeDlg::FormShow(TObject *Sender)
{
   CenterDialog(this,AppMainForm);
   Timer1->Enabled = true;
}
//---------------------------------------------------------------------------
void __fastcall TWipeDlg::Timer1Timer(TObject *Sender)
{
   // this function only returns status of last file handled
   Timer1->Enabled = false;
   if (files->Count)
   {
      if (FileExists(files->Strings[0]))
      {
         AnsiString file = files->Strings[0];
         Label1->Caption = "...wiping "+ExtractFileName(file)+"...";
         AppProcessMessages();
         retcode = simplewipe(file,passes);
      }
      files->Delete(0);
      Timer1->Enabled = true;
      return;
   }
   ModalResult = mrOk;
}
//---------------------------------------------------------------------------
int __fastcall GetWipeBuffer(struct isaac_state *state, int pass, int buffer[256],
   int *patterns)
{
   if (pass<NUMPATTERNS) filltable(state,buffer,patterns[pass]);
   else filltable(state,buffer,-1);
   return 0;
}
//---------------------------------------------------------------------------
AnsiString __fastcall GetWipeErrorStr()
{
   DWORD result = GetLastError();
   LPVOID lpMsgBuf;
   FormatMessage(
      FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
      NULL,
      result,
      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
     (LPTSTR) &lpMsgBuf,
      0,
      NULL
   );
   return AnsiString((char*)lpMsgBuf);
}

// The following Bob Jenkins' ISAAC generator
// (http://ourworld.compuserve.com/homepages/bob_jenkins/isaacafa.htm)
// which is designed to be cryptographically strong and very fast.
// I'm not sure if I'd trust it for key material, but it should be
// strong enough for this application, and the speed is nice when
// you need megabytes of output.

// This index operation is more efficient on many processors
#define ind(mm,x) *(unsigned*) ((char*)(mm)+((x) & ((ISAAC_SIZE-1)<<2)))

// The central step. This uses two temporaries, x and y
#define isaac_step(mix,a,b,mm,m,off,r)    \
(                                         \
   a = (a^(mix)) + m[off],                \
   x = *m,                                \
   *(m++) = y = ind(mm,x) + a + b,        \
   *(r++) = b = ind(mm,y>>ISAAC_LOG) + x  \
)

// The basic seed-scrambling step
#define mix(a,b,c,d,e,f,g,h)      \
(                                 \
   a ^= b << 11, d += a,          \
   b += c, b ^= c >>  2, e += b,  \
   c += d, c ^= d <<  8, f += c,  \
   d += e, d ^= e >> 16, g += d,  \
   e += f, e ^= f << 10, h += e,  \
   f += g, f ^= g >>  4, a += f,  \
   g += h, g ^= h <<  8, b += g,  \
   h += a, h ^= a >>  9, c += h,  \
   a += b                         \
)

// Refill the entire r array with one call.
static void isaac_refill(struct isaac_state *s, int r[ISAAC_SIZE])
{
	register int *m=s->mm; // Pointer into state array
	register int x,y; // Temps needed by isaac_step() macro
	register int a,b; // Caches of a and b

	a = s->aa;
	b = s->bb + (++s->cc);

	do
	{
		isaac_step(a << 13, a, b, s->mm, m, +ISAAC_SIZE/2, r);
		isaac_step(a >>  6, a, b, s->mm, m, +ISAAC_SIZE/2, r);
		isaac_step(a <<  2, a, b, s->mm, m, +ISAAC_SIZE/2, r);
		isaac_step(a >> 16, a, b, s->mm, m, +ISAAC_SIZE/2, r);
	} while (m < s->mm+ISAAC_SIZE/2);

	do
	{
		isaac_step(a << 13, a, b, s->mm, m, -ISAAC_SIZE/2, r);
		isaac_step(a >>  6, a, b, s->mm, m, -ISAAC_SIZE/2, r);
		isaac_step(a <<  2, a, b, s->mm, m, -ISAAC_SIZE/2, r);
		isaac_step(a >> 16, a, b, s->mm, m, -ISAAC_SIZE/2, r);
	} while (m < s->mm+ISAAC_SIZE);

	s->bb = b;
	s->aa = a;
}

// Initialize the ISAAC RNG with the given 256-entry seed array.
// That array may be the mm[] array.
static void isaac_init(struct isaac_state *s, int const *seed)
{
 	 int i,a,b,c,d,e,f,g,h,passes=(seed!=0);
#if 0
	 // The original presentation
	 a=b=c=d=e=f=g=h=0x9e3779b9; // the golden ratio
	 for (i = 0; i < 4; ++i)
      mix(a,b,c,d,e,f,g,h); // scramble it
#else
 	 // A smaller and faster, although more opaque version.
   a = 0x1367df5a;
	 b = 0x95d90059;
	 c = 0xc3163e4b;
	 d = 0x0f421ad8;
	 e = 0xd92a4a78;
	 f = 0xa51a3c49;
	 g = 0xc4efea1b;
	 h = 0x30609119;
#endif
	 do
	 {
    // Fill up mm[] with messy stuff
	    for (i = 0; i < ISAAC_SIZE; i += 8)
		 {
		  	 if (seed)
			 {
			  	 // use all the information in the seed
				 a += seed[i];
				 b += seed[i+1];
				 c += seed[i+2];
				 d += seed[i+3];
				 e += seed[i+4];
				 f += seed[i+5];
				 g += seed[i+6];
				 h += seed[i+7];
			 }
			 mix(a, b, c, d, e, f, g, h);
			 s->mm[i] = a;
			 s->mm[i+1] = b;
			 s->mm[i+2] = c;
			 s->mm[i+3] = d;
			 s->mm[i+4] = e;
			 s->mm[i+5] = f;
			 s->mm[i+6] = g;
			 s->mm[i+7] = h;
		 }
		 seed = s->mm;
		 // If we were given a seed, do a second pass to make
		 // all of the seed affect all of mm
	} while (passes--);

	s->aa = s->bb = s->cc = 0;
}

// Seed the RNG, using 160 bits of seed material, which should be *plenty*.
// This cheats and uses the s->mm[] array to hold the seed material,
// which is allowed by issac_init().
static void isaac_seed(struct isaac_state *state)
{
//  Get some good seed material.
    RAND_bytes((byte*)state->mm,20);  // <-- Got this func from SSLLib
	isaac_init(state, state->mm);
}
// End RNG code

// Shuffle an array of ints into random order.
// This is done by swapping the last element in the array
// with a random element of the array and then repeating
// for the array without the last element.
static void shuffle(int *array, int num, struct isaac_state *s)
{
	int i,t,r[ISAAC_SIZE];

	isaac_refill(s,r);
	while (--num)
	{
		i = (int)(r[num] % (num+1));
		t = array[i];
		array[i] = array[num];
		array[num] = t;
	}
}

// Move the random pass into the middle of the shuffle
static void tweakshuffle(int *array, int num)
{
	for (int i = 0; i < num; i++)
	{
		if (array[i] < 0)
		{
			num /= 2;
			array[i] = array[num];
			array[num] = -1;
			break;
		}
	}
}

// Pass types:
// 0xXXX - Repeat XXX to fill file
// -1 - Random
static void filltable(struct isaac_state *state, int buf[ISAAC_SIZE], int type)
{
	if (type<0)
		isaac_refill(state,(int*)buf);
	else
	{
		((unsigned char*)buf)[0] = (char)(255 & (type >> 4));
		((unsigned char*)buf)[1] = (char)(255 & ((type << 4) | (type >> 8)));
		((unsigned char*)buf)[2] = (char)(255 & type);

     int i;
		for (i=3; i<ISAAC_BYTES/2; i*=2)
			memcpy((char*)buf+i, (char*)buf, i);
		memcpy((char*)buf+i, (char*)buf, ISAAC_BYTES-i);
	}
}
//---------------------------------------------------------------------------






